function [lambda_p, lambda_m, X, info] = somddpa(M, E, K, B, C, s0, opts)
%SOMDDPA Second-order modally-damped dominant pole algorithm.
%
% SYNTAX:
%   [lambda_p, lambda_m, X, info] = SOMDDPA(M, E, K, B, C, s0)
%   [lambda_p, lambda_m, X, info] = SOMDDPA(M, E, K, B, C, s0, opts)
%
% DESCRIPTION:
%   This function implements a version of the subspace accelerated
%   quadratic MIMO dominant pole algorithm for modally damped second-order
%   systems
%
%       Mx''(t) + Ex'(t) + Kx(t) = Bu(t),                               (1)
%                           y(t) = Cx(t),                               (2)
%
%   with the assumptions M, E, K symmetric positive definite and
%   E*inv(M)*K = K*inv(M)*E. It computes poles of the transfer function
%
%       H(s) = C * inv(s^2 * M + s * E + K) * B
%
%   that are dominant with respect to a certain measure. The poles are
%   computed in pairs lambda_p, lambda_m such that
%
%       lambda_p = a + b, lambda_m = a - b
%
%   are eigenvalues of (s^2 * M + s * E + K)X = 0, with X containing the
%   corresponding real eigenvectors.
%
% INPUT:
%   M    - matrix of dimensions n x n from (1), symmetric positive definite
%   E    - matrix of dimensions n x n from (1), symmetric positive definite
%   K    - matrix of dimensions n x n from (1), symmetric positive definite
%   B    - matrix of dimensions n x m from (1)
%   C    - matrix of dimensions p x n from (2)
%   s0   - scalar, initial guess of a dominant pole
%   opts - options struct, with the following entries
%   +-----------------+---------------------------------------------------+
%   |    PARAMETER    |                     MEANING                       |
%   +-----------------+---------------------------------------------------+
%   | DPStrategy      | string, the measure of dominance for the poles    |
%   |                 |  'LS'   - smallest absolute value                 |
%   |                 |  'LR'   - smallest absolute real part             |
%   |                 |  'LP'   - smallest absolute product               |
%   |                 |  'LRP'  - smallest absolute product of real parts |
%   |                 |  'LRPI' - as LRP but restriction to complex       |
%   |                 |           conjugate poles                         |
%   |                 | (default 'LRP')                                   |
%   +-----------------+---------------------------------------------------+
%   | Expansion       | string, way of expanding the search space         |
%   |                 |  'real'    - RMGS on real parts of vectors        |
%   |                 |  'real_qr' - QR decomp. on real parts of vectors  |
%   |                 |  'full'    - RMGS on real and imaginary parts of  |
%   |                 |              vectors                              |
%   |                 |  'full_qr' - QR decomp. on real and imaginary     |
%   |                 |              parts of vectors                     |
%   |                 | (default 'full_qr')                               |
%   +-----------------+---------------------------------------------------+
%   | Kmax            | positive integer, maximum dimension of search     |
%   |                 | space before restarting                           |
%   |                 | (default 30)                                      |
%   +-----------------+---------------------------------------------------+
%   | Kmin            | positive integer, dimension of search space after |
%   |                 | restarting                                        |
%   |                 | (default 10)                                      |
%   +-----------------+---------------------------------------------------+
%   | Linearization   | {1, 2, 3}, determining type of underlying         |
%   |                 | linearization                                     |
%   |                 |  1 - K-modified first companion form              |
%   |                 |  2 - first companion form                         |
%   |                 |  3 - second companion form                        |
%   |                 | (default 3)                                       |
%   +-----------------+---------------------------------------------------+
%   | MaxIter         | positive integer, maximum number of iteration     |
%   |                 | steps                                             |
%   |                 | (default 100)                                     |
%   +-----------------+---------------------------------------------------+
%   | Nwanted         | positive integer, number of requested pole pairs  |
%   |                 | (default 5)                                       |
%   +-----------------+---------------------------------------------------+
%   | Output          | {0, 1, 2}, flag for amount of printed iteration   |
%   |                 | information                                       |
%   |                 |  0 - no print out                                 |
%   |                 |  1 - found pole pairs are shown                   |
%   |                 |  2 - information about each step                  |
%   |                 | (default 0)                                       |
%   +-----------------+---------------------------------------------------+
%   | Reortho         | {0, 1}, turns off/on the reorthogonalization of   |
%   |                 | search space after found pole pair or restart     |
%   |                 | (default 1)                                       |
%   +-----------------+---------------------------------------------------+
%   | Target          | scalar, reorders dominance measure to get poles   |
%   |                 | close to the target                               |
%   |                 | (default [])                                      |
%   +-----------------+---------------------------------------------------+
%   | TDefl           | {0, 1, 2}, flag for turbo deflation               |
%   |                 |  0 - classical deflation                          |
%   |                 |  1 - turbo deflation using the first-order        |
%   |                 |      realization                                  |
%   |                 |  2 - turbo delfation using the second-order       |
%   |                 |      realization (experimental)                   |
%   |                 | (default 1)                                       |
%   +-----------------+---------------------------------------------------+
%   | Tolerance       | positive scalar, tolerance for residual of        |
%   |                 | computed poles and eigenvectors                   |
%   |                 | (default 1.0e-08)                                 |
%   +-----------------+---------------------------------------------------+
%
% OUTPUT:
%	lambda_p - vector, with plus parts of found poles
%   lambda_m - vector, with minus parts of found poles
%   X        - real matrix, containing the corresponding eigenvectors
%   info     - struct, containing information about the iteration
%   +-----------------+---------------------------------------------------+
%   |      ENTRY      |                     MEANING                       |
%   +-----------------+---------------------------------------------------+
%   | Dominance       | vector, with the resulting dominance values of    |
%   |                 | the eigenvalue pairs                              |
%   +-----------------+---------------------------------------------------+
%   | History         | vector, with the iteration steps when the pole    |
%   |                 | pairs were found                                  |
%   +-----------------+---------------------------------------------------+
%   | NormRes         | vector, with the residuals of the used chosen     |
%   |                 | Ritz value in every iteration step                |
%   +-----------------+---------------------------------------------------+
%   | Residues        | cell array, with the residual matrices from the   |
%   |                 | series expansion of the transfer function         |
%   +-----------------+---------------------------------------------------+
%
%
% REFERENCES:
%   On dominant pole algorithms for second-order systems:
%   [1] J. Rommes, N. Martins, Computing transfer function dominant poles
%       of large-scale second-order dynamical systems, IEEE Trans. Power
%       Syst. 21 (4) (2006) 1471--1483 (Nov. 2006).
%       doi:10.1109/TPWRS.2006.881154
%   [2] P. Benner, P. Krschner, Z. Tomljanovic, N. Truhar, Semi-active
%       damping optimization of vibrational systems using the parametric
%       dominant pole algorithm, Z. Angew. Math. Mech. 96 (5) (2016)
%       604--619 (2016).
%       doi:10.1002/zamm.201400158
%   On the pole-residual formulation of modally-damped systems:
%   [3] C. A. Beattie, P. Benner, H2-optimality conditions for structured
%       dynamical systems, Preprint MPIMD/14-18, Max Planck Institute
%       Magdeburg (2014).
%       URL https://www2.mpi-magdeburg.mpg.de/preprints/2014/18/

%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU Affero General Public License as published
% by the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU Affero General Public License for more details.
%
% You should have received a copy of the GNU Affero General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%
% Copyright (C) 2019 Peter Benner, Steffen W. R. Werner
%


if exist('OCTAVE_VERSION', 'builtin')
    warn_state = warning('off', 'Octave:nearly-singular-matrix');
else
    warn_state = warning('off', 'MATLAB:nearlySingularMatrix');
end


%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% CHECK OPTIONS.                                                          %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if nargin < 7
    opts = struct([]) ;
end

% Size of the problem.
[n, ~] = size(M);
n2     = 2 * n;

% Number of wanted eigentriplets.
if isfield(opts, 'Nwanted') && ~isempty(opts.Nwanted)
    nwanted = opts.Nwanted;
else
    nwanted = 5;
end

% Flag for turning on output informations.
if isfield(opts,'Output') && ~isempty(opts.Output)
    output = opts.Output;
else
    output = 0;
end

% Tolerance for accuracy of Ritz values.
if isfield(opts, 'Tolerance') && ~isempty(opts.Tolerance)
    tol = opts.Tolerance;
else
    tol = 1e-08;
end

% Set type of linearization.
if isfield(opts, 'Linearization') && ~isempty(opts.Linearization)
    lin = opts.Linearization;
else
    lin = 3;
end

% Maximum number of iteration steps.
if isfield(opts, 'MaxIter') && ~isempty(opts.MaxIter)
    maxiter = opts.MaxIter;
else
    maxiter = 100;
end

% Minimal dimension of the subspace after restart.
if isfield(opts, 'Kmin') && ~isempty(opts.Kmin)
    kmin = opts.Kmin;
else
    kmin = 10;
end

% Maximal dimension of the subspace before restart.
if isfield(opts, 'Kmax') && ~isempty(opts.Kmax)
    kmax = opts.Kmax;
else
    kmax = 30;
end

% Choose strategy for searching of dominant poles.
if isfield(opts, 'DPStrategy') && ~isempty(opts.DPStrategy)
    dpstrat = opts.DPStrategy;
else
    dpstrat = 'LRP';
end

% Influence ordering by giving a target eigenvalue to be close.
if isfield(opts, 'Target') && ~isempty(opts.Target)
    target = opts.Target;
else
    target = [];
end

% Choose type of deflation.
if isfield(opts, 'TDefl') && ~isempty(opts.TDefl)
    tdefl = opts.TDefl;
else
    tdefl = 1;
end

% Basis reorthogonalization.
if isfield(opts, 'Reortho') && ~isempty(opts.Reortho)
    reortho = opts.Reortho;
else
    reortho = 1;
end

if isfield(opts, 'Expansion') && ~isempty(opts.Expansion)
    expansion = opts.Expansion;
else
    expansion = 'full_qr';
end


%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% INITIALIZATION.                                                         %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

lambda_p = zeros(nwanted, 1); % found eigenvalues (positive part)
lambda_m = zeros(nwanted, 1); % found eigenvalues (negative part)

% Choose first-order formulation for turbo deflation.
switch lin
    case 1
        % [ -K  0 ] [ x  ]'   [  0  -K ] [ x  ]   [ 0 ]
        % [       ] [    ]  = [        ] [    ] + [   ] u,
        % [  0  M ] [ x' ]    [ -K  -E ] [ x' ]   [ B ]
        %
        %                                [ x  ]
        %             y     = [  C   0 ] [    ]
        %                                [ x' ]
        Eaug = blkdiag(-K, M);
    case 2
        % [ I  0 ] [ x  ]'   [  0   I ] [ x  ]   [ 0 ]
        % [      ] [    ]  = [        ] [    ] + [   ] u,
        % [ 0  M ] [ x' ]    [ -K  -E ] [ x' ]   [ B ]
        %
        %                               [ x  ]
        %             y    = [  C   0 ] [    ]
        %                               [ x' ]
        Eaug = blkdiag(speye(n), M);
    case 3
        % [ E  M ] [ x  ]'   [ -K  0 ] [ x  ]   [ B ]
        % [      ] [    ]  = [       ] [    ] + [   ] u,
        % [ M  0 ] [ x' ]    [  0  M ] [ x' ]   [ 0 ]
        %
        %                              [ x  ]
        %            y     = [ C   0 ] [    ]
        %                              [ x' ]
        Eaug = [E, M; M, sparse(n,n)];
end

% Matrices of found eigenvectors.
Xfound      = zeros(n, 0);
Xfound_aug  = zeros(n2, 0);
EXfound_aug = zeros(n2, 0);

normres = [];

% Initialize subspace matrices.
V = zeros(n, 0);

% Initialize iteration info vector.
hist = [];

% Initialize iteration variables.
imagtol   = 1e-8; % relative tolerance for imaginary part of real values
iter      = 0;    % iteration counter
st        = s0;   % current chosen shift
converged = 0;    % convergence flag

% Print table layout for display mode.
if output > 1
    fprintf('iter  dim(V)       theta               normr_r\n');
    fprintf('----------------------------------------------\n');
end

% Get size of inputs and Outputs.
m = size(B, 2);
p = size(C, 1);

% Flag for square transfer function matrices.
issquare = (m == p);

ep_opts.b       = B;
ep_opts.c       = C';
ep_opts.dpstrat = dpstrat;
ep_opts.target  = target;
ep_opts.tdefl   = tdefl;


%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Precomputation                                                          %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Compute initial transfer function matrix.
RB = (st^2 * M + st * E + K) \ B;
H  = C * RB;

% If square use eigenvalues else singular values.
if issquare
    [u, d] = eigs(H , 1, 'LM', struct('disp', 0, 'v0', ones(m, 1)));
    [y, ~] = eigs(H', 1, 'LM', struct('disp', 0, 'v0', ones(p, 1)));
    y      = y / (u' * y);
else
    [y, d, u] = svd(full(H));
    
    y = y(:, 1);
    d = d(1, 1);
    u = u(:, 1);
end

% Build up augmented matrices for turbo-deflation.
if tdefl == 1
    C_d = [C'; zeros(n, p)];
    if lin == 3
        B_d = [B; zeros(n, m)];
    else
        B_d = [zeros(n, m); B];
        
        [LK, UK] = lu(K);
        BK_d     = (UK \ (LK \ B_d(1:n, :)));
        CK_d     = (LK' \ (UK' \ C_d(1:n, :)));
    end
end

% Turbo deflation using the second order structure.
if tdefl == 2
    ep_opts.E = E;
    ep_opts.M = M;
end


%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Iteration.                                                              %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

while iter < maxiter
    iter = iter + 1;
    
    %%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Solve SPD linear systems.                                           %
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    if tdefl ~= 1 % conventionel deflation
        Q = st^2 * M + st * E + K;
        s = Q \ (B * u);
        t = Q' \ (C' * y);
        
        newton_upd = st - d / (t' * (2 * st * (M * s) + E * s));
        
        if ~tdefl
            tmp = [s; newton_upd * s] ;
            tmp = rmgs_two(tmp, Xfound_aug, EXfound_aug, 1, 1);
            s   = tmp(1:n, 1);
            
            tmp = [t ; newton_upd' * t];
            tmp = rmgs_two(tmp, Xfound_aug, EXfound_aug, 1, 1);
            t   = tmp(1:n, 1);
        end
    else % deflation via i/o vectors of linearized problem
        Q = st^2 * M + st * E + K;
        s = Q \ ((st * B_d(n+1:end, :) + B_d(1:n, :)) * u);
        t = Q' \ ((st' * C_d(n+1:end, :) + C_d(1:n, :)) * y);
        
        if lin ~= 3
            s = (-BK_d * u + s) / st;
            t = (-CK_d * y + t) / st';
        end
    end
    
    
    %%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Enlarge projection basis.                                           %
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    k = size(V, 2);
    
    switch expansion
        case 'full'
            kj = k;
            
            for w = {real(s), imag(s), real(t), imag(t)}
                [v, h] = rmgs(w{1}, V, 1);
                if (h > 0) && (kj < n)
                    V(:, kj+1) = v / h;
                    kj         = kj + 1;
                end
            end
            
        case 'full_qr'
            [V(:, 1:k+4), ~] = qr([V(:, 1:k), real(s), imag(s), ...
                real(t), imag(s)], 0);
            
        case 'real'
            kj = k;
            
            for w = {real(s), real(t)}
                [v, h] = rmgs(w{1}, V, 1);
                if (h > 0) && (kj < n)
                    V(:, kj+1) = v / h;
                    kj         = kj + 1;
                end
            end
            
        case 'real_qr'
            [V(:, 1:k+2), ~] = qr([V(:, 1:k), real(s), real(t)], 0);
            
        otherwise
            error('SOMDDPA:wrongExpansion', ...
                'The requested expansion method does not exist!');
    end
    
    dk = size(V, 2) - k;
    
    assert(dk > 0, ...
        'SOMDDPA:stagnation', ...
        'Stagnation detected!');
    
    %%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Construct reduced order szstem.                                     %
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    Mr = V' * (M * V);
    Er = V' * (E * V);
    Kr = V' * (K * V);
    
    k = k + dk;
    
    
    %%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Eigenvalue selection.                                               %
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    [Lp, Lm, X] = mdqep(Mr, Er, Kr);
    
    if tdefl == 1
        ep_opts.b = B_d;
        ep_opts.c = C_d;
    elseif tdefl == 2
        QXfound   = st^2 * M * Xfound + st * E * Xfound + K * Xfound;
        ep_opts.b = B - QXfound * ((Xfound' * QXfound) \ (Xfound' * B));
        ep_opts.c = C' - QXfound * ((Xfound' * QXfound)' \ (Xfound' * C'));
    end
    
    % Ordering eigenvalues corresponding to dominance measure.
    [Lp, Lm, X] = order_qdp(Lp, Lm, X, V, ep_opts);
    
    found     = 1;
    firstloop = 1;
    
    
    %%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Iteration on dominant pairs.                                        %
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    while found
        v     = V * X(:, 1);
        v     = v / norm(v, 2);
        theta = Lp(1);
        
        st = Lp(1);
        nr = norm(st^2 * M * v + st * E * v + K * v, 2);
        
        if firstloop
            normres = [normres, nr]; %#ok<AGROW>
        end
        
        % Print current approximation of dominant pole pair.
        if output > 1
            fprintf('%4d | %4d | %+10.5g %+10.5gi | %.1e\n', ...
                iter, size(V, 2), real(theta), imag(theta), nr);
        end
        
        
        %%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Processing of approximations.                                   %
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        found = (nr < tol);
        if found
            %%
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % Found dominant pole pair.                                   %
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            
            hist = [hist; iter]; %#ok<AGROW>
            
            if abs(imag(Lp(1))) / abs(Lp(1)) < imagtol % Real pole pair.
                converged = converged + 1;
                
                lambda_p(converged) = real(Lp(1));
                lambda_m(converged) = real(Lm(1));
                
                % Print found real dominant pole pair.
                if output > 0
                    fprintf('real pair %d found at iter %4d: %s, %s\n', ...
                        converged, iter, ...
                        num2str(lambda_p(converged)), ...
                        num2str(lambda_m(converged)));
                end
                
                Xfound     = [Xfound, v]; %#ok<AGROW>
                Xfound_aug = [Xfound_aug, [v; lambda_p(converged) * v], ...
                    [v; lambda_m(converged) * v]]; %#ok<AGROW>
            else % Complex conjugate pole pair.
                converged = converged + 1;
                
                lambda_p(converged) = Lp(1);
                lambda_m(converged) = Lm(1);
                
                % Print found complex conjugate dominant pole pair.
                if output > 0
                    fprintf(['compl. conj. pair %d found at ' ...
                        'iter %4d: %s, %s\n'], ...
                        converged, iter, ...
                        num2str(lambda_p(converged)), ...
                        num2str(lambda_m(converged)));
                end
                
                Xfound     = [Xfound, v]; %#ok<AGROW>
                Xfound_aug = [Xfound_aug, [v; real(Lp(1)) * v], ...
                    [v; imag(Lp(1)) * v]]; %#ok<AGROW>
            end
            
            
            %%
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % Deflation.                                                  %
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            EXfound_aug = [EXfound_aug, ...
                Eaug * Xfound_aug(:, end-1:end)]; %#ok<AGROW>
            nc          = Xfound_aug(:, end-1:end)' ...
                * EXfound_aug(:, end-1:end);
            EXfound_aug(:, end-1:end) = EXfound_aug(:,end-1:end) / nc;
            
            % Remove found eigenvalue and corresponding vectors.
            kr = 2:k;
            Lp = Lp(kr);
            Lm = Lm(kr);
            V  = V * X(:, kr);
            k  = size(V, 2);
            
            if tdefl == 1  % Deflate i/o vectors of linearized problem.
                B_d = B_d - EXfound_aug * (Xfound_aug' * B_d);
                C_d = C_d - EXfound_aug * (Xfound_aug' * C_d);
                
                if lin ~= 3
                    BK_d = (UK \ (LK \ B_d(1:n, :)));
                    CK_d = (LK' \ (UK' \ C_d(1:n, :)));
                end
            end
            
            % Reorthogonalization.
            if reortho
                kk = 1;
                
                while kk <= k
                    if isreal(Lp(kk))
                        augv = [V(:, kk); Lp(kk) * V(:, kk)];
                        augv = rmgs_two(augv, Xfound_aug, ...
                            EXfound_aug, 1, 1);
                        
                        [V(:, kk), vv] = rmgs(augv(1:n), V(:, 1:kk-1), 1);
                        V(:, kk)       = V(:, kk) / vv(end);
                    else
                        augv = [V(:, kk); (real(Lp(kk)) ...
                            - imag(Lp(kk))) * V(:, kk)];
                        augv = rmgs_two(augv, Xfound_aug, ...
                            EXfound_aug, 1, 1);
                        
                        [V(:, kk), vv] = rmgs(augv(1:n), V(:, 1:kk-1), 1);
                        V(:, kk)       = V(:, kk) / vv(end);
                    end
                    
                    kk = kk + 1;
                end
            end
            
            % Reprojection.
            Mr = V' * (M * V);
            Er = V' * (E * V);
            Kr = V' * (K * V);
            
            
            %%
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % Break condition.                                            %
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            if converged >= nwanted
                break;
            end
            
            
            %%
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % Select further poles.                                       %
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            found = (k > 0);
            if found
                [Lp, Lm, X] = mdqep(Mr, Er, Kr);
                
                if tdefl == 1
                    ep_opts.b = B_d;
                    ep_opts.c = C_d;
                elseif tdefl == 2
                    QXfound   = st^2 * M * Xfound ...
                        + st * E * Xfound + K * Xfound;
                    ep_opts.b = B - QXfound * ((Xfound' * QXfound) ...
                        \ (Xfound' * B));
                    ep_opts.c = C' - QXfound * ((Xfound' * QXfound)' ...
                        \ (Xfound' * C'));
                end
                
                % Ordering eigenvalues corresponding to dominance measure.
                [Lp, Lm, X] = order_qdp(Lp, Lm, X, V, ep_opts);
            end
            
            
            %%
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % Update direction vectors.                                   %
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            if (converged < nwanted) && ~found
                RB = (st^2 * M + st * E + K) \ B;
                H  = C * RB;
                
                if issquare
                    [u, d] = eigs(H , 1, 'LM', ...
                        struct('disp', 0, 'v0', ones(m, 1)));
                    [y, ~] = eigs(H', 1, 'LM', ...
                        struct('disp', 0, 'v0', ones(p, 1)));
                    y      = y / (u' * y);
                else
                    [y, d, u] = svd(full(H));
                    
                    y = y(:, 1);
                    d = d(1, 1);
                    u = u(:, 1);
                end
            end
            
        elseif k >= kmax
            %%
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % Restart method.                                             %
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            fprintf('Restart.\n');
            
            % Compute residuals.
            k  = kmin;
            kr = 1:kmin;
            l  = length(Lp);
            Ld = zeros(2 * l, 2 * l);
            Xt = zeros(l, 2 * l);
            
            for h = 1:l
                Ld(2 * h - 1:2 * h, 2 * h - 1:2 * h) = ...
                    [real(Lp(h)), imag(Lp(h)); imag(Lm(h)), real(Lm(h))];
                Xt(:, 2 * h - 1:2 * h) = [X(:, h), X(:, h)];
            end
            
            RR = M * (V * (Xt * Ld^2)) ...
                + E * (V * (Xt * Ld)) ...
                + K * (V * Xt);
            
            % Get minimal residual.
            nrs      = num2cell(RR, 1);
            nrs      = cellfun(@(x) norm(x, 2), nrs);
            [~, idx] = min(nrs);
            
            if mod(idx, 2)
                idx = (idx + 1) / 2;
            else
                idx = idx / 2;
            end
            
            if idx >= kmin
                kr = [idx, 1:(kmin - 1)];
            end
            
            % Truncate smaller basis.
            Lp = Lp(kr);
            Lm = Lm(kr);
            V  = V * X(:, kr);
            
            % Reorthogonalization.
            if reortho
                kk = 1;
                
                while kk <= k
                    if isreal(Lp(kk))
                        augv = [V(:, kk); Lp(kk) * V(:, kk)];
                        augv = rmgs_two(augv, Xfound_aug, ...
                            EXfound_aug, 1, 1);
                        
                        [V(:, kk), vv] = rmgs(augv(1:n), V(:, 1:kk-1), 1);
                        V(:, kk)       = V(:, kk) / vv(end);
                    else
                        augv = [V(:, kk); (real(Lp(kk)) ...
                            - imag(Lp(kk))) * V(:, kk)];
                        augv = rmgs_two(augv, Xfound_aug, ...
                            EXfound_aug, 1, 1);
                        
                        [V(:, kk), vv] = rmgs(augv(1:n), V(:, 1:kk-1), 1);
                        V(:, kk)       = V(:, kk) / vv(end);
                    end
                    
                    kk = kk + 1;
                end
            end
            
            % Update direction vectors.
            RB = (st^2 * M + st * E + K) \ B;
            H  = C * RB;
            
            if issquare
                [u, d] = eigs(H , 1, 'LM', ...
                    struct('disp', 0, 'v0', ones(m, 1)));
                [y, ~] = eigs(H', 1, 'LM', ...
                    struct('disp', 0, 'v0', ones(p, 1)));
                y      = y / (u' * y);
            else
                [y, d, u] = svd(full(H));
                
                y = y(:, 1);
                d = d(1, 1);
                u = u(:, 1);
            end
        end
        
        firstloop = 0;
    end
    
    
    %%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Formatting output.                                                  %
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    if (converged >= nwanted) || (iter >= maxiter)
        
        if iter >= maxiter
            warning('SOMDDPA:convergence', ...
                'No convergence in %d iterations.', ...
                iter);
        end
        
        lambda_p = lambda_p(1:converged);
        lambda_m = lambda_m(1:converged);
        residues = cell(converged, 1);
        nres     = zeros(converged, 1);
        
        % Compute residual terms.
        for i = 1:converged
            RM          = (C * Xfound_aug(1:n, i)) ...
                * (Xfound_aug(1:n,i).' * B ...
                * sqrt(abs(lambda_p(i) * lambda_m(i))));
            residues{i} = RM;
            nres(i)     = norm(RM, 2);
        end
        
        % Sort with respect to dominance.
        switch dpstrat
            case 'LS'
                nres = nres ./ abs(lambda_p);
            case 'LR'
                nres = nres ./ abs(real(lambda_p));
            case 'LP'
                nres = nres ./ abs(lambda_p .* lambda_m);
            case 'LRP'
                nres = nres ./ abs(real(lambda_p) .* real(lambda_m));
            case 'LRPI'
                nres = (nres .* abs(imag(lambda_p))) ...
                    ./ abs(real(lambda_p) .* real(lambda_m));
            otherwise
                error('SOMDDPA:dominance', ...
                    'Not implemented measure of dominance.');
        end
        
        [nres, idx] = sort(nres, 'descend');
        
        residues = residues(idx);
        Xfound   = Xfound(:, idx);
        
        % Formatting of output.
        X = Xfound;
        
        info = struct( ...
            'Dominance', nres, ...
            'History'  , hist, ...
            'NormRes'  , normres, ...
            'Residues' , {residues});
        
        break;
    end
end

warning(warn_state);

end


%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% LOCAL FUNCTION: MDQEP                                                   %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [Ds, D2s, Xs] = mdqep(Mr, Er, Kr)
%MDQEP Modally-Damped eigenproblem solver.
%
% DESCRIPTION:
%   This function computes the pairwise eigenvalues of the modally-damped
%   eigenproblem and the corresponding real eigenvectors such that
%
%       (Ds^2 * Mr + Ds * Er + Kr) * Xs   = 0,
%       (D2s^2 * Mr + D2s * Er + Kr) * Xs = 0.
%
% INPUT:
%   Mr  - matrix of dimensions r x r
%   Er  - matrix of dimensions r x r
%   Kr  - matrix of dimensions r x r
%
% OUTPUT:
%   Ds  - vector of length r, plus parts of eigenvalues
%   D2s - vector of length r, minus parts of eigenvalues
%   Xs  - matrix of dimensions r x r, real eigenvectors

n   = size(Mr, 1);
Ds  = zeros(n, 1);
D2s = zeros(n, 1);

% Underlying generalized eigenvalue problem.
[Xs, Omega2] = eig(Kr, Mr);
Omega        = sqrt(diag(Omega2));

% Scaling of eigenvectors.
Xs = Xs * diag(1./sqrt(diag(Xs' * Mr * Xs)));
Xs = Xs * diag(1./sqrt(Omega));

% Computing the damping coefficients.
Xi = zeros(n, 1);
for i = 1:n
    Xi(i) = 1/2 * Xs(:, i)' * (Er * Xs(:, i));
end

% Computing the second-order eigenvalue pairs.
for k = 1:n
    if abs(Xi(k) - 1) < 1.0e-09
        Ds(k)  = -Omega(k);
        D2s(k) = -Omega(k);
        
        warning('MDSAMDPA:eig', ...
            'Multiple eigenvalue found %e', Ds(k));
    else
        Ds(k)  = -Xi(k) * Omega(k) + Omega(k) ...
            * sqrt((Xi(k) + 1) * (Xi(k) - 1));
        D2s(k) = -Xi(k) * Omega(k) - Omega(k) ...
            * sqrt((Xi(k) + 1) * (Xi(k) - 1));
    end
end

end


%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% LOCAL FUNCTION: ORDER_QDP                                               %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [D, Dt, Xt] = order_qdp(D, Dt, Xt, V, opts)
%ORDER_QDP Ordering of eigenvalue pairs.
%
% DESCRIPTION:
%   This function implements different strategies for ordering the
%   eigenvalue pairs with respect to certain dominance measures.
%
% INPUT:
%   D    - vector of length r, plus part of eigenvalue pairs
%   Dt   - vector of length r, minus part of eigenvalues pairs
%   Xt   - matrix of dimensions r x r, real eigenvectors
%   V    - matrix of dimensions n x r, projection matrix
%   opts - option struct, containing the following entries
%   +-----------------+---------------------------------------------------+
%   |    PARAMETER    |                     MEANING                       |
%   +-----------------+---------------------------------------------------+
%   | E               | damping matrix of the original system,            |
%   |                 | only needed if tdefl == 2                         |
%   +-----------------+---------------------------------------------------+
%   | M               | mass matrix of the original system,               |
%   |                 | only needed if tdefl == 2                         |
%   +-----------------+---------------------------------------------------+
%   | tdefl           | {0, 1, 2}, flag for turbo deflation               |
%   |                 |  0 - classical deflation                          |
%   |                 |  1 - turbo deflation using the first-order        |
%   |                 |      realization                                  |
%   |                 |  2 - turbo delfation using the second-order       |
%   |                 |      realization (experimental)                   |
%   +-----------------+---------------------------------------------------+
%   | dpstrat         | string, the measure of dominance for the poles    |
%   |                 |  'LS'   - smallest absolute value                 |
%   |                 |  'LR'   - smallest absolute real part             |
%   |                 |  'LP'   - smallest absolute product               |
%   |                 |  'LRP'  - smallest absolute product of real parts |
%   |                 |  'LRPI' - as LRP but restriction to complex       |
%   |                 |           conjugate poles                         |
%   +-----------------+---------------------------------------------------+
%   | target          | scalar, reorders dominance measure to get poles   |
%   |                 | close to the target                               |
%   +-----------------+---------------------------------------------------+
%   | b               | input matrix, has to fit to the chosen            |
%   |                 | linearization and turbo deflation strategy        |
%   +-----------------+---------------------------------------------------+
%   | c               | output matrix, has to fit to the chosen           |
%   |                 | linearization and turbo deflation strategy        |
%   +-----------------+---------------------------------------------------+
%
% OUTPUT:
%   D    - reorderd vector of length r, plus part of eigenvalue pairs
%   Dt   - reorderd vector of length r, minus part of eigenvalues pairs
%   Xt   - reorderd matrix of dimensions r x r, real eigenvectors


%% Initialization.
B       = opts.b;
C       = opts.c';
n       = length(D);
X       = V * Xt;
residue = zeros(n, 1);

if opts.tdefl == 2
    E = opts.E;
    M = opts.M;
end


%% Compute residuals.
for j = 1:n
    if opts.tdefl==1
        Xaug = [X(:,j); D(j)*X(:,j)];
    end
    
    omega = sqrt(real(D(j) * Dt(j)));
    
    if opts.tdefl == 1
        residue(j) = norm(omega * (C * Xaug) * (Xaug' * B), 2);
    elseif opts.tdefl == 2
        residue(j) = norm(omega * (C * X(:,j)) * (X(:,j)' * B), 2) ...
            / abs(X(:,j)' * (2 * D(j) * M * X(:,j) + E * X(:,j)));
    else
        residue(j) = norm(omega * (C * X(:, j)) * (X(:,j)' * B), 2);
    end
    
    switch opts.dpstrat
        case 'LS'
            residue(j) = residue(j) / abs(D(j));
        case 'LR'
            residue(j) = residue(j) / abs(real(D(j)));
        case 'LP'
            residue(j) = residue(j) / abs(D(j) * Dt(j));
        case 'LRP'
            residue(j) = residue(j) / abs(real(D(j)) * real(Dt(j)));
        case 'LRPI'
            residue(j) = (residue(j) * abs(imag(D(j)))) ...
                / abs(real(D(j)) * real(Dt(j)));
        otherwise
            error('MDSAMDPA:dominance', ...
                'Not implemented version of dominance.');
    end
end


%% Sort residuals and eigenvalues/eigenvectors.
if ~isempty(opts)
    if ~isempty(opts.dpstrat)
        if isempty(opts.target) % default ordering
            [~, idx] = sort(residue, 'descend');
        else % target ordering
            g = 0.95;
            
            [~, idx] = sort(g * residue + (1 - g) ...
                ./ abs(D - opts.target)', 'descend');
        end
        D  = D(idx);
        Dt = Dt(idx);
        Xt = Xt(:, idx);
    end
end

end


%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% LOCAL FUNCTION: RMGS                                                    %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [w, h] = rmgs(v, V, ref)
%RMGS Repeated modified Gram-Schmidt method.
%
% DESCRIPTION:
%   This function orthogonalizes the given vector v with respect to V.
%
% INPUT:
%   v   - vector of length n
%   V   - orthogonal matrix of dimensions n x k
%   ref - {0, 1} turns refinement step off/on
%
% OUTPUT:
%   w   - orthogonalized vector
%   h   - norm of w

if nargin < 3
    ref = 0 ;
end

[~, k] = size(V) ;

if k < 1
    w = v;
    h = norm(w, 2);
    return;
end

w     = v;
nrmv  = norm(v, 2);
kappa = 0.1 ;

for j = 1:k
    w = w - (V(:, j)' * w) * V(:, j) ;
end

if ref && (norm(w, 2) < kappa * nrmv)
    for j = 1:k
        w = w - (V(:, j)' * w) * V(:, j) ;
    end
end

h = norm(w, 2);

end


%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% LOCAL FUNCTION: RMGS_TWO                                                %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function v = rmgs_two(v, Q, Qt, ref, ogs)
%RMGS_TWO Repeated modified Gram-Schmidt for skew projection.
%
% DESCRIPTION:
%   This function uses the repeated modified Gram-Schmidt method for
%   computing the skew projection (I - Q * Qt') * v.
%
% INPUT:
%   v   - vector of length n
%   Q   - bi-orthogonal matrix of dimensions n x k such that Qt'*Q = I
%   Qt  - bi-orthogonal matrix of dimensions n x k such that Qt'*Q = I
%   ref - {0, 1} turns refinement step off/on if ogs == 0
%         (default 0)
%   ogs - {0, 1} switches between modified Gram-Schmidt (0) and direct
%         projection (1)
%         (default 0)
%
% OUTPUT:
%   v   - projected vector

if nargin < 4
    ref = 0;
    ogs = 0;
end

if nargin < 5
    ogs = 0;
end

[n,  k]  = size(Q);
[nt, kt] = size(Qt);

assert(isequal([n, k], [nt, kt]), ...
    'SOMDDPA:rmgs_two', ...
    'Unequal basis dimensions!');

if k < 1
    return;
end

nrmv  = norm(v, 2);
kappa = 0.1 ;

if ogs
    v = v - Q * (Qt' * v);
    v = v - Q * (Qt' * v);
    return;
end

for j = 1:k
    v = v - (Qt(:, j)' * v) * Q(:, j);
end

if ref && (norm(v, 2) < kappa * nrmv)
    for j = 1 : k
        v = v - (Qt(:, j)' * v) * Q(:, j);
    end
end

end
